#include <conio.h>
#include <stdio.h>
#include <memory.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

struct vector2f {
	float	x,y;
};

struct vector3f {
	float	x,y,z;
};

struct ogf_bbox {
	struct vector3f		min;
	struct vector3f		max;
};

struct ogf_bsphere {
	struct vector3f		c;
	float			r;
};

struct ChunkIFF
{
	unsigned int type, size;
};

struct OGFHeader_v4 // 0x2C (44)
{
 unsigned char format_version;
 unsigned char type;
 unsigned short shader_id;
 struct ogf_bbox		bb;
 struct ogf_bsphere	bs;
};

struct OGFVertex // 32 (0x20)
{
	struct vector3f p; // point
	struct vector3f n; // normal
	float u,v;
};

char destFlag;
char * lineptr2;

#define WHITESPACE " \t\n\r"

#define MAX_V_COUNT 65530
#define MAX_I_COUNT 60000
#define MAX_MTL_COUNT 500
#define MAX_MSH_COUNT 17000

unsigned long meshcount = 0;
unsigned long normalcount = 0;
unsigned long uvcount = 0;

unsigned long verticescount;
unsigned long indicescount;

struct mesh
{
OGFVertex * vertices;
unsigned long vtxcount;
unsigned short * indices;
unsigned long idxcount;
char * m_name;
char * m_dds;
char * m_shd;
};

unsigned short materialcount = 0;

struct material
{
	unsigned short id; 
	char * m_name;
	char * txt_name0;
	char * txt_name1;
};

material * mtl;
mesh * msh;
vector2f * uvcoord;

#define VEC_Cross(a, b, c)((c).x = (a).y * (b).z - (a).z * (b).y, (c).y = (a).z * (b).x - (a).x * (b).z, (c).z = (a).x * (b).y - (a).y * (b).x)
#define VEC_Subtract(a, b, c)((c).x = (a).x - (b).x, (c).y = (a).y - (b).y, (c).z = (a).z - (b).z)
#define VEC_Magnitude(v)(sqrt((v).x * (v).x + (v).y * (v).y + (v).z * (v).z))
#define VEC_Normalizef(v){ float __mag = 1.0f / (float)VEC_Magnitude(v); \
(v).x *= __mag; (v).y *= __mag; (v).z *= __mag; }
void Make_Normal(vector3f a, vector3f b, vector3f c, vector3f * n)
{
	vector3f v1 = {0,0,0},v2 = {0,0,0};
	VEC_Subtract(b, a, v1);
	VEC_Subtract(c, a, v2);
	VEC_Cross(v1, v2, *n);
	VEC_Normalizef(*n);
}

/////////////////////////////////////////////////////////

void parsemtl(char* name)
{
	char objbuf2[256];
	char mtlName[256];
  FILE *mtlFile;
	char curblk2 = 0, lastblk2 = 0;

  strcpy(mtlName,name);
  strcpy(strrchr(mtlName,'.'),".mtl");
  mtlFile = fopen(mtlName,"rb");

printf("---begin(%s)---\n",mtlName);
for (;;)
{
lineptr2 = (char*)&objbuf2;
if (!fgets(objbuf2,255,mtlFile)) break;
// printf("\t\t\t_DEBUG_READ %s",objbuf2);
if (strstr(objbuf2,"newmtl")) 
	{
  curblk2 = 1; if  (curblk2 != lastblk2) printf("BLK_MATS\n");
	materialcount++;

  lineptr2 = strtok(lineptr2,WHITESPACE);
	lineptr2 = strtok(NULL,WHITESPACE);
  mtl[materialcount-1].m_name = new char[strlen(lineptr2)+1];
  strcpy(mtl[materialcount-1].m_name,lineptr2);
	printf(" MATERIAL[%u] = '%s'\n",materialcount-1,mtl[materialcount-1].m_name);
  mtl[materialcount-1].txt_name0 = NULL;
  mtl[materialcount-1].txt_name1 = NULL;
  lastblk2 = curblk2;
	}
else
if (strstr(objbuf2,"map_Ka ")) 
	{
  curblk2 = 2;
  lineptr2 = strtok(lineptr2,WHITESPACE);
	lineptr2 = strtok(NULL,WHITESPACE);
	char * s = strchr(lineptr2,'.'); if (s) lineptr2[s-lineptr2] = 0x00;
  mtl[materialcount-1].txt_name0 = new char[strlen(lineptr2)+1];
  strcpy(mtl[materialcount-1].txt_name0,lineptr2);
	printf(" MAP_KA[%u] = '%s'\n",materialcount-1,mtl[materialcount-1].txt_name0);
  lastblk2 = curblk2;
	}
else
if (strstr(objbuf2,"map_Kd ")) 
	{
  curblk2 = 3;
  lineptr2 = strtok(lineptr2,WHITESPACE);
	lineptr2 = strtok(NULL,WHITESPACE);
	char * s = strchr(lineptr2,'.'); if (s) lineptr2[s-lineptr2] = 0x00;
  mtl[materialcount-1].txt_name1 = new char[strlen(lineptr2)+1];
  strcpy(mtl[materialcount-1].txt_name1,lineptr2);
	printf(" MAP_KD[%u] = '%s'\n",materialcount-1,mtl[materialcount-1].txt_name1);
  lastblk2 = curblk2;
	}
/*
else
	printf("unk\n");
*/
}
printf("---end---\n");

  fclose(mtlFile);
}

char curblk = 0, lastblk = 0;
char objbuf[255];
char * lineptr;

int main(int argc, char* argv[])
{
	unsigned long i,j;
	char fileName[256], fileName2[256];
  FILE *inputFile, *outputFile;
	
	if (argc < 2) 
  {
		printf("Wavefront obj to X-Ray OGF (OBJ2OGF) parser/converter [201101xx]\n\n");
		printf("usage: OBJ2OGF srcfile.obj [ destfile.ogf ]\nif destfile.ogf is omitted - only dump will be provided\n");
		return 1;
	}

	strcpy((char*)fileName,argv[1]);

  inputFile = fopen(fileName,"rb");
	if(!inputFile)
  {
		printf("X-Ray OBJ2OGF: cannot open file for read.\n");
		return 1;
  }
	if (argc == 2) 
	{
		printf("X-Ray OBJ2OGF: output to screen.\n");
	} 
else 
  {
	destFlag = 1;
		strcpy((char*)fileName2,argv[2]);
		outputFile = fopen(fileName2,"wb");
		if(!outputFile)
		{
			printf("OGF: cannot open file for write.\n");
			return 0;
		}
	}

/////////////////////////////////////////////////////////


mtl = new material[MAX_MTL_COUNT];
msh = new mesh[MAX_MSH_COUNT];

printf("---begin(%s)---\n",fileName);

parsemtl(fileName);

for (;;)
{
lineptr = (char*)&objbuf;
if (!fgets(objbuf,255,inputFile)) break;
// printf("\t\t\t_DEBUG_READ %s",objbuf);
if (strstr(lineptr,"mtllib")) 
	{  if (!materialcount) { printf("mtllib().parse\n"); parsemtl(fileName); } }
else
if (strstr(objbuf,"v ")) 
{
  curblk = 1; if  (curblk != lastblk) 
  {
 	meshcount++; verticescount = normalcount = uvcount = 0;
	msh[meshcount-1].vertices = new OGFVertex[MAX_V_COUNT];
	printf("BLK_VERTICES[%u]\n",meshcount-1);
	}
  lineptr = strtok(lineptr,WHITESPACE);
  lineptr = strtok(NULL,WHITESPACE); msh[meshcount-1].vertices[verticescount].p.x = (float)atof(lineptr);
  lineptr = strtok(NULL,WHITESPACE); msh[meshcount-1].vertices[verticescount].p.y = (float)atof(lineptr);
  lineptr = strtok(NULL,WHITESPACE); msh[meshcount-1].vertices[verticescount].p.z = (float)atof(lineptr);

 // fix for orion v1

  msh[meshcount-1].vertices[verticescount].p.x /= 100 / 0.93384f;
  msh[meshcount-1].vertices[verticescount].p.y /= 100 / 0.70017f;
  msh[meshcount-1].vertices[verticescount].p.z /= 100;

/*
	printf(" -> mesh[%u].v[%u] = (%f,%f,%f)\n", meshcount-1, verticescount,
	msh[meshcount-1].vertices[verticescount].p.x, 
	msh[meshcount-1].vertices[verticescount].p.y,
	msh[meshcount-1].vertices[verticescount].p.z);
*/
  verticescount++; if (verticescount == MAX_V_COUNT) {printf("vlimit reached\n");_getch();};
  msh[meshcount-1].vtxcount = verticescount;
  lastblk = curblk;
}
else
if (strstr(objbuf,"vn ")) 
{
  curblk = 2; if  (curblk != lastblk) 
  {
	printf("BLK_NORMALES[%u]\n",meshcount-1);
	}
  lineptr = strtok(lineptr,WHITESPACE);
  lineptr = strtok(NULL,WHITESPACE); msh[meshcount-1].vertices[normalcount].n.x = (float)atof(lineptr);
  lineptr = strtok(NULL,WHITESPACE); msh[meshcount-1].vertices[normalcount].n.y = (float)atof(lineptr);
  lineptr = strtok(NULL,WHITESPACE); msh[meshcount-1].vertices[normalcount].n.z = (float)atof(lineptr);
/*
	printf(" -> mesh[%u].n[%u] = (%f,%f,%f)\n", meshcount-1, normalcount,
	msh[meshcount-1].vertices[normalcount].n.x, 
	msh[meshcount-1].vertices[normalcount].n.y,
	msh[meshcount-1].vertices[normalcount].n.z);
*/  
	normalcount++; if (normalcount == MAX_V_COUNT) {printf("vlimit reached\n");_getch();};
  lastblk = curblk;
}
else
if (strstr(objbuf,"vt ")) 
{
  curblk = 3; if  (curblk != lastblk) 
  {
  uvcoord = new vector2f[MAX_V_COUNT];
	printf("BLK_UVCOORDS[%u]\n",meshcount-1);
	}
  lineptr = strtok(lineptr,WHITESPACE);
  lineptr = strtok(NULL,WHITESPACE); uvcoord[uvcount].x = (float)atof(lineptr);
  lineptr = strtok(NULL,WHITESPACE); uvcoord[uvcount].y = (float)atof(lineptr);
/*
	printf(" -> mesh[%u].uv[%u] = (%f,%f)\n", meshcount-1, uvcount,
	uvcoord[uvcount].x,
	uvcoord[uvcount].y);
*/
  uvcount++; if (uvcount == MAX_V_COUNT) {printf("vlimit reached\n");_getch();};
  lastblk = curblk;
}
else
if (strstr(objbuf,"f ")) 
{
  long idx;
  unsigned long na,nb,nc;
  vector3f n;

  curblk = 4; if  (curblk != lastblk) 
 	{
	indicescount = 0;
	msh[meshcount-1].indices = new unsigned short[MAX_I_COUNT];
	printf("BLK_INDICES[%u]\n",meshcount-1);
	}
  lineptr = strtok(lineptr,WHITESPACE);
  lineptr = strtok(NULL,"/"); idx = atoi(lineptr); if (idx < 0) {msh[meshcount-1].indices[indicescount] = msh[meshcount-1].vtxcount+idx;} else msh[meshcount-1].indices[indicescount] = idx-1;

  na = msh[meshcount-1].indices[indicescount];
  msh[meshcount-1].vertices[msh[meshcount-1].indices[indicescount]].u = uvcoord[indicescount].x,
  msh[meshcount-1].vertices[msh[meshcount-1].indices[indicescount]].v = 1.0f-uvcoord[indicescount].y;

//printf(" -> mesh[%u].idx[%u] = (%u,", meshcount-1, indicescount, msh[meshcount-1].indices[indicescount]); 
	indicescount++;
  if (normalcount) lineptr = strtok(NULL,"/"); 
  lineptr = strtok(NULL," \t");
  lineptr = strtok(NULL,"/"); idx = atoi(lineptr); if (idx < 0) {msh[meshcount-1].indices[indicescount] = msh[meshcount-1].vtxcount+idx;} else msh[meshcount-1].indices[indicescount] = idx-1;

  nb = msh[meshcount-1].indices[indicescount];
  msh[meshcount-1].vertices[msh[meshcount-1].indices[indicescount]].u = uvcoord[indicescount].x,
  msh[meshcount-1].vertices[msh[meshcount-1].indices[indicescount]].v = 1.0f-uvcoord[indicescount].y;

//printf("%u,", msh[meshcount-1].indices[indicescount]); 
	indicescount++;
  if (normalcount) lineptr = strtok(NULL,"/"); 
  lineptr = strtok(NULL," \t"); 
  lineptr = strtok(NULL,"/"); idx = atoi(lineptr); if (idx < 0) {msh[meshcount-1].indices[indicescount] = msh[meshcount-1].vtxcount+idx;} else msh[meshcount-1].indices[indicescount] = idx-1;

  nc = msh[meshcount-1].indices[indicescount];
  msh[meshcount-1].vertices[msh[meshcount-1].indices[indicescount]].u = uvcoord[indicescount].x,
  msh[meshcount-1].vertices[msh[meshcount-1].indices[indicescount]].v = 1.0f-uvcoord[indicescount].y;

  Make_Normal(
	msh[meshcount-1].vertices[na].p,
	msh[meshcount-1].vertices[nb].p,
	msh[meshcount-1].vertices[nc].p,&n);
  msh[meshcount-1].vertices[na].n = n;
  msh[meshcount-1].vertices[nb].n = n;
  msh[meshcount-1].vertices[nc].n = n;

//printf("%u)\n", msh[meshcount-1].indices[indicescount]); 
	indicescount++;
  if (normalcount) lineptr = strtok(NULL,"/"); 
  lineptr = strtok(NULL,WHITESPACE);
	if (indicescount == MAX_I_COUNT) {printf("ilimit reached\n");_getch();};
  msh[meshcount-1].idxcount = indicescount;
  lastblk = curblk;
}
else
if (strstr(objbuf,"usemtl")) 
	{
  curblk = 5; if  (curblk != lastblk) printf("BLK_MATID[%u]\n",meshcount-1);
  lineptr = strtok(lineptr,WHITESPACE);
	lineptr = strtok(NULL,WHITESPACE);
  msh[meshcount-1].m_name = new char[strlen(lineptr)+1];
  strcpy(msh[meshcount-1].m_name,lineptr);
	printf(" MATERIAL[%u] = '%s'\n",meshcount-1,msh[meshcount-1].m_name);
  lastblk = curblk;
	}
/*
else
if (strstr(objbuf,"g ")) 
	printf("g data\n");
else
if (strstr(objbuf,"o")) 
	printf("o data\n");
else
	printf("unk\n");
*/
}
printf("---end---\n");

//delete [] vertices;
//delete [] indices;

  printf("wavefront obj parse complete\n");

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////


  for (j = 0; j < meshcount; j++)
  for (i = 0; i < materialcount; i++)
  {
  //printf("(msh) '%s' <-> '%s' (mtl)\n",msh[j].m_name,mtl[i].m_name);
  if (!strcmp(msh[j].m_name,mtl[i].m_name)) 
  {
	msh[j].m_shd = new char[10]; strcpy(msh[j].m_shd,"default");

	if (mtl[i].txt_name0 && strlen(mtl[i].txt_name0))
  {msh[j].m_dds = new char[strlen(mtl[i].txt_name0)+1], strcpy(msh[j].m_dds,mtl[i].txt_name0); }
  else 
	if (mtl[i].txt_name1 && strlen(mtl[i].txt_name1))
  {msh[j].m_dds = new char[strlen(mtl[i].txt_name1)+1]; strcpy(msh[j].m_dds,mtl[i].txt_name1);}
  else
	msh[j].m_dds = new char[10], strcpy(msh[j].m_dds,"default"); 

  printf("->> mesh[%u] mat[%u]: dds,shd - '%s','%s' v:%u i:%u\n",j,i,msh[j].m_dds,msh[j].m_shd,msh[j].vtxcount,msh[j].idxcount);
  }

  }

  printf("mesh <-> material match complete\n");

if (argc >= 3)
{
  unsigned int tot_i = 0, tot_v = 0;

  printf("creating '%s'\n",fileName2);
  outputFile = fopen(fileName2,"wb");

  ChunkIFF * ih = new ChunkIFF;
  ih->type = 0x01; // header
  ih->size = sizeof(OGFHeader_v4);
	fwrite(ih,sizeof(ChunkIFF),1,outputFile);

  OGFHeader_v4 * hdr4 = new OGFHeader_v4;
  hdr4->format_version = 4;
  hdr4->type = 1;
  hdr4->shader_id = 0;
  hdr4->bb.min.x = 
  hdr4->bb.min.y = 
  hdr4->bb.min.z = 
  hdr4->bb.max.x = 
  hdr4->bb.max.y = 
  hdr4->bb.max.z = 0.0f;
  // zero fill :)
  hdr4->bs.r = 0.0f; 
  hdr4->bs.c.x = 0.0f;
  hdr4->bs.c.y = 0.0f;
  hdr4->bs.c.z = 0.0f;
	fwrite(hdr4,sizeof(OGFHeader_v4),1,outputFile);

  ih->type = 0x12; // text info
  ih->size = 0x6E; // guessed :)
	fwrite(ih,sizeof(ChunkIFF),1,outputFile);
  i = 0; 
  fputs("test_level\\test_level.object",outputFile);
  fwrite(&i,1,1,outputFile); 
  fputs("\\\\MSHOME\\Administrator",outputFile); fwrite(&i,1,1,outputFile); 
  fwrite(&i,4,1,outputFile); // fake time
  fputs("\\\\MSHOME\\Administrator",outputFile); fwrite(&i,1,1,outputFile); 
  fwrite(&i,4,1,outputFile); // fake time
  fputs("\\\\MSHOME\\Administrator",outputFile); fwrite(&i,1,1,outputFile); 
  fwrite(&i,4,1,outputFile); // fake time


  ih->type = 0x09; // childrens
  ih->size = 8 + (8+sizeof(OGFHeader_v4))+(0x14+8)+(8+((msh[0].vtxcount*(sizeof(OGFVertex)))+8))+(8+((msh[0].idxcount*2)+4));
	fwrite(ih,sizeof(ChunkIFF),1,outputFile);

  printf("making %u blocks\n",meshcount);

	for (j = 0;j < meshcount; j++)
	{

	//printf("child[%u] - v:%u i:%u\n",j,msh[j].vtxcount,msh[j].idxcount);

	// child start

  ih->type = j; 
  ih->size = (8+sizeof(OGFHeader_v4))+((strlen(msh[j].m_dds) + strlen(msh[j].m_shd) + 0x02)+8)+(8+((msh[j].vtxcount*(sizeof(OGFVertex)))+8))+(8+((msh[j].idxcount*2)+4));

	fwrite(ih,sizeof(ChunkIFF),1,outputFile);

	// .... 1.2.3.4

  ih->type = 0x01; // header
  ih->size = sizeof(OGFHeader_v4);
	fwrite(ih,sizeof(ChunkIFF),1,outputFile);
  hdr4->format_version = 4;
  hdr4->type = 0;
  hdr4->shader_id = 0;
	fwrite(hdr4,sizeof(OGFHeader_v4),1,outputFile);

  ih->type = 0x02; // texture
  ih->size = strlen(msh[j].m_dds) + strlen(msh[j].m_shd) + 0x02; 
	fwrite(ih,sizeof(ChunkIFF),1,outputFile);
	i = 0;
  fputs(msh[j].m_dds,outputFile); fwrite(&i,1,1,outputFile); 
  fputs(msh[j].m_shd,outputFile); fwrite(&i,1,1,outputFile); 

	tot_v += msh[j].vtxcount;
  ih->type = 0x03; // verts
  ih->size = (msh[j].vtxcount*(sizeof(OGFVertex)))+8;
	fwrite(ih,sizeof(ChunkIFF),1,outputFile);
	i = 0x00000112; // D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_TEX1
  fwrite(&i,4,1,outputFile);
	i = msh[j].vtxcount; 
  fwrite(&i,4,1,outputFile);
  for (unsigned int s = 0; s < i ; s++)
	fwrite(&msh[j].vertices[s],sizeof(OGFVertex),1,outputFile);
	
	tot_i += msh[j].idxcount;
  ih->type = 0x04; // idx
  ih->size = (msh[j].idxcount*2)+4;
	fwrite(ih,sizeof(ChunkIFF),1,outputFile);
	i = msh[j].idxcount; 
  fwrite(&i,4,1,outputFile);
	fwrite(msh[j].indices,sizeof(unsigned short)*i,1,outputFile);

	printf("."); if ((j % 50)==0) printf("\r%u",j);
	// child end 

	}

  delete hdr4;
  delete ih;
	if (inputFile) fclose(inputFile);
  if (outputFile) fclose(outputFile);
  printf("\nwriting '%s' [v %u, i %u] OK\n",fileName2,tot_v,tot_i);
}

  for (j = 0; j < meshcount; j++)
  delete msh[j].vertices,
  delete msh[j].indices,
  delete msh[j].m_name,
  delete msh[j].m_dds,
  delete msh[j].m_shd;
  delete [] msh;

  for (i = 0; i < materialcount; i++)
  delete mtl[i].m_name,
  delete mtl[i].txt_name0,
  delete mtl[i].txt_name1;
  delete [] mtl;

  delete [] uvcoord;

	return 0;
}